from __future__ import annotations

import asyncio
from contextlib import suppress
from typing import List
from urllib.parse import urldefrag

from lxml.etree import ParserError
from lxml.html import document_fromstring

from .utils import URL_RE_MULTI, threaded


class GenericSource:
    def __init__(
        self,
        location: str,
        content: str,
        depth: int = 0,
    ):
        self.location = location
        self.content = content
        self.depth = depth
        self.loop = asyncio.get_running_loop()

    @staticmethod
    def is_valid(url: str) -> bool:
        if not url.startswith("http"):
            return False
        return True

    @threaded
    def find_urls(self) -> List[str]:
        return [
            urldefrag(url)[0]
            for url in URL_RE_MULTI.findall(self.content)
            if self.is_valid(urldefrag(url)[0])
        ]


class FileSource(GenericSource):
    @classmethod
    def from_path(cls, path: str) -> FileSource:
        with open(path) as f:
            return cls(location=path, content=f.read())


class HTMLSource(GenericSource):
    @threaded
    def find_urls(self) -> List[str]:
        with suppress(ValueError, ParserError):
            tree = document_fromstring(self.content)
            tree.make_links_absolute(self.location)
            return [
                urldefrag(url[2])[0]
                for url in tree.iterlinks()
                if self.is_valid(urldefrag(url[2])[0])
            ]
        return []



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List
from urllib.parse import urlparse


class URLFilter(ABC):
    @abstractmethod
    def validate(self, url: str) -> bool:
        pass

    def apply(self, urls: List[str]) -> List[str]:
        return [url for url in urls if self.validate(url)]


class OrFilter(List[URLFilter], URLFilter):
    def validate(self, url: str) -> bool:
        return not self or any(url_filter.validate(url) for url_filter in self)


class DomainFilter(URLFilter):
    def __init__(self, domain: str):
        if domain.startswith("www."):
            domain = domain[4:]
        self.domain = domain

    def validate(self, url: str) -> bool:
        return urlparse(url).netloc in [self.domain, f"www.{self.domain}"]

    @classmethod
    def from_url(cls, url: str) -> DomainFilter:
        return cls(domain=urlparse(url).netloc)



import asyncio
import re
from functools import partial, wraps
from pathlib import Path
from typing import Any, Awaitable, Callable, List, TypeVar


T = TypeVar("T")

URL_RE = "https?:\\/\\/(?:www\\.)?[-a-zA-Z0-9@:%._\\+~#=]{1,256}\\.[a-zA-Z0-9()]{1,6}\\b(?:[-a-zA-Z0-9()@:%_\\+.~#?&\\/=]*)"  # noqa
URL_RE_SINGLE = re.compile(f"^{URL_RE}$")
URL_RE_MULTI = re.compile(URL_RE)


def is_url(value: str) -> bool:
    return bool(URL_RE_SINGLE.match(value))


def truncate_url(url: str, max_length: int, placeholder: str) -> str:
    if len(url) > max_length:
        url = url[: max_length - len(placeholder)] + placeholder
    return url


def get_files_path(path: str, recursive: bool = False) -> List[str]:
    root_path = Path(path).resolve()
    files_path, glob = [], ""
    while not root_path.exists():
        root_path, glob = (
            root_path.parent,
            (Path(root_path.name) / glob).as_posix(),
        )

    if root_path.is_dir():
        for file_path in root_path.glob(glob or ("**/*" if recursive else "*")):
            if file_path.is_file():
                files_path.append(file_path)
    elif root_path.is_file():
        files_path.append(root_path)

    return [file_path.as_posix() for file_path in files_path]


def threaded(func: Callable[..., T]) -> Callable[..., Awaitable[T]]:
    @wraps(func)
    def _wrapped(*args: Any, **kwargs: Any) -> Any:
        loop = asyncio.get_running_loop()
        return loop.run_in_executor(None, partial(func, *args, **kwargs))

    return _wrapped



from typing import Dict, List, Optional

from .node import Node
from .request import Request, RequestState


class Store:
    def __init__(self):
        self._requests: Dict[str, Request] = {}
        self._graph: Dict[str, Node] = {}

    def add_request(self, request: Request) -> None:
        self._requests[request.url] = request

    def create_or_get_node(self, location: str) -> Node:
        node = self._graph.get(location)
        if not node:
            node = Node(location=location, request=self._requests.get(location))
            self._graph[location] = node
        return node

    def add_edge(self, location_a: str, location_b: str) -> None:
        node_a = self.create_or_get_node(location_a)
        node_b = self.create_or_get_node(location_b)
        node_a.links.append(node_b)

    def get_request(self, url: str) -> Optional[Request]:
        return self._requests.get(url)

    def get_requests_error(self) -> List[Request]:
        return [
            request
            for request in self._requests.values()
            if request.state is RequestState.COMPLETED and not request.ok()
        ]

    def get_requests_completed(self) -> List[Request]:
        return [
            request
            for request in self._requests.values()
            if request.state is RequestState.COMPLETED
        ]

    def get_nodes_processed(self) -> List[Node]:
        return [node for node in self._graph.values() if node.links_all_completed()]



import asyncio
from contextlib import asynccontextmanager
from dataclasses import dataclass
from signal import SIGINT, SIGTERM
from typing import Any, AsyncGenerator, Iterable, List, Optional, Set, Type, Union

from aiolimiter import AsyncLimiter

from .engine import ENGINE_MAPPING, Engine
from .filter import DomainFilter, OrFilter
from .node import Node
from .output import Output
from .request import Request, RequestState
from .source import FileSource, GenericSource, HTMLSource
from .store import Store
from .utils import get_files_path, is_url


DEFAULT_CONCURRENCY = 20
DEFAULT_RECURSION_MODE = "internal"
DEFAULT_CHECK_MODE = "all"
DEFAULT_DEPTH = -1
DEFAULT_TIMEOUT = 20
DEFAULT_ENGINE = "aiohttp"
DEFAULT_MAX_RETRY = 1
DEFAULT_RATE_LIMIT = 20


@dataclass
class CrawlerResult:
    requests: List[Request]
    nodes: List[Node]


class Crawler:
    def __init__(
        self,
        concurrency: int = DEFAULT_CONCURRENCY,
        recursion_mode: str = DEFAULT_RECURSION_MODE,
        check_mode: str = DEFAULT_CHECK_MODE,
        depth: int = DEFAULT_DEPTH,
        timeout: int = DEFAULT_TIMEOUT,
        max_retry: int = DEFAULT_MAX_RETRY,
        engine: Union[str, Type[Engine]] = DEFAULT_ENGINE,
        rate_limit: int = DEFAULT_RATE_LIMIT,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        self.loop = loop or asyncio.get_event_loop()
        self.concurrency = concurrency
        self.depth = depth
        self.timeout = timeout
        self.max_retry = max_retry
        self.recursion_mode = recursion_mode
        self.recursion_filter = OrFilter([])
        self.check_mode = check_mode
        self.check_filter = OrFilter([])
        self.store = Store()
        self.output = Output()
        self.requests_queue: asyncio.Queue[Request] = asyncio.Queue()
        self.sources_queue: asyncio.Queue[GenericSource] = asyncio.Queue()
        self.tasks: Set[asyncio.Task[Any]] = set()
        self._stopping: bool = False
        self._limiter = AsyncLimiter(max_rate=rate_limit, time_period=1)

        if isinstance(engine, type) and issubclass(engine, Engine):
            self.engine_class = engine
        elif isinstance(engine, str):
            self.engine_class = ENGINE_MAPPING.get(engine)  # type: ignore

        if self.engine_class is None:
            raise Exception(f"Cannot instanciate Crawler with engine: {engine}")

    def start(self, sources: Iterable[str]) -> CrawlerResult:
        for signal in (SIGINT, SIGTERM):
            self.loop.add_signal_handler(
                signal, lambda: asyncio.create_task(self.stop())
            )

        try:
            self.loop.run_until_complete(self._async_start(sources))
        finally:
            self.display_result()

        return CrawlerResult(
            requests=self.store.get_requests_completed(),
            nodes=self.store.get_nodes_processed(),
        )

    async def _async_start(self, sources: Iterable[str]) -> None:

        # Initialize startup sources
        for source in sources:
            if is_url(source):
                if self.recursion_mode == "internal":
                    self.recursion_filter.append(DomainFilter.from_url(source))
                if self.check_mode == "internal":
                    self.check_filter.append(DomainFilter.from_url(source))
                self.add_request(url=source)
            else:
                for file_path in get_files_path(source):
                    file_source = FileSource.from_path(file_path)
                    await self.add_source(file_source)

        # Initialize workers
        async with self.engine_class() as engine:
            request_workers = [
                self._request_worker(engine) for _ in range(self.concurrency)
            ]
            self.tasks = {
                asyncio.create_task(coro)
                for coro in [
                    *request_workers,
                    self.source_worker(),
                    self.output.start(),
                    self.requests_queue.join(),
                ]
            }
            # Here we wait for the requests manager to have no more requests
            # to process and for other tasks to capture inner tasks
            # exceptions different from asyncio.CancelledError
            while not (self.requests_queue.empty() and self.sources_queue.empty()):
                done, pending = await asyncio.wait(
                    self.tasks, return_when=asyncio.FIRST_COMPLETED
                )
                if any(task.cancelled() for task in done):
                    break

            for task in done:
                if not task.cancelled():
                    exc = task.exception()
                    if exc:
                        raise exc
            if not self._stopping:
                await self.stop(tasks=pending)

    async def stop(
        self, tasks: Optional[Set[asyncio.Task]] = None  # type: ignore
    ) -> None:
        self._stopping = True
        tasks = tasks or self.tasks
        if tasks:
            for task in tasks:
                task.cancel()
            await asyncio.gather(*tasks, return_exceptions=True)

    async def add_source(self, source: GenericSource) -> None:
        source_node = self.store.create_or_get_node(location=source.location)
        for url in await source.find_urls():
            if self.check_filter.validate(url):
                self.add_request(url=url, depth=source.depth)
                self.store.add_edge(source_node.location, url)
        self.output.notify_node(source_node)

    def add_request(
        self,
        url: str,
        depth: int = 0,
    ) -> None:
        request = self.store.get_request(url=url)
        if not request:
            request = Request(
                url=url, depth=depth, timeout=self.timeout, max_retry=self.max_retry
            )
            self.store.add_request(request)
            self.requests_queue.put_nowait(request)
            self.output.notify_request(request)

    @asynccontextmanager
    async def _request_context(self, request: Request) -> AsyncGenerator[None, None]:
        async with self._limiter:
            self._set_request_state(request, RequestState.IN_PROGRESS)
            yield
            self._set_request_state(request, RequestState.COMPLETED)

    def _set_request_state(self, request: Request, request_state: RequestState) -> None:
        request.state = request_state
        self.output.notify_request(request)

    async def _request_worker(self, engine: Engine) -> None:
        while True:
            request = await self.requests_queue.get()
            async with self._request_context(request):
                response = await request.fetch(engine)
            if (
                self.recursion_filter.validate(request.url)
                and response
                and response.text
                and response.ok()
                and (self.depth < 0 or request.depth < self.depth)
            ):
                source = HTMLSource(
                    content=response.text, location=request.url, depth=request.depth + 1
                )

                if (
                    self.requests_queue.qsize() < 2 * self.concurrency
                    and self.sources_queue.empty()
                ):
                    await self.add_source(source)
                else:
                    self.sources_queue.put_nowait(source)

            self.requests_queue.task_done()

    async def source_worker(self):
        while True:
            if self.requests_queue.qsize() < 2 * self.concurrency:
                source = await self.sources_queue.get()
                await self.add_source(source)
                self.sources_queue.task_done()
            else:
                await asyncio.sleep(0.05)

    def display_result(self):
        # print(self.store.count_requests_by_state(), end="\x1b[0K\n")
        error_requests = self.store.get_requests_error()
        if error_requests:
            print("Failed requests:", end="\x1b[0K\n")
            for request in error_requests:
                print(request.result(), end="\x1b[0K\n")



from __future__ import annotations

from typing import List, Optional

from .request import Request, RequestState
from .utils import truncate_url


class Node:
    def __init__(
        self,
        location: str,
        request: Optional[Request],
    ):
        self.location = location
        self.request = request
        self.links: List[Node] = []

    def count_links_total(self) -> int:
        return len(self.links)

    def count_links_ok(self) -> int:
        return len([link for link in self.links if link.request and link.request.ok()])

    def links_all_completed(self) -> bool:
        return all(
            [
                link.request.state is RequestState.COMPLETED
                for link in self.links
                if link.request
            ]
        )

    def count_links_completed(self):
        return len(
            [
                link
                for link in self.links
                if link.request and link.request.state is RequestState.COMPLETED
            ]
        )

    def count_in_progress(self) -> int:
        return len(
            [
                link
                for link in self.links
                if link.request and link.request.state is RequestState.IN_PROGRESS
            ]
        )

    def get_result_string(self) -> str:
        return (
            f"{truncate_url(self.location, 79, '...')} "
            f"[{self.count_links_ok()}/{self.count_links_total()}] "
        )



from sys import version_info

if version_info >= (3, 8):
    from importlib import metadata as metadata  # type: ignore
else:
    import importlib_metadata as metadata  # type: ignore

__version__ = metadata.version(__package__)



from __future__ import annotations

import asyncio
from enum import Enum
from typing import TYPE_CHECKING, Optional

from .response import Response
from .utils import truncate_url

if TYPE_CHECKING:
    from .engine import Engine


class RequestState(Enum):
    NOT_STARTED = 0
    IN_PROGRESS = 1
    COMPLETED = 2


class RequestError(Exception):
    pass


class RequestTimeoutError(RequestError):
    def __str__(self) -> str:
        return "Request timeout"


class RequestConnectionError(RequestError):
    def __str__(self) -> str:
        return "Connection error"


class Request:
    def __init__(self, url: str, max_retry: int = 1, timeout: int = 20, depth: int = 0):
        self.url = url
        self.max_retry = max_retry
        self.timeout = timeout
        self.state = RequestState.NOT_STARTED
        self.depth = depth
        self.response: Optional[Response] = None
        self.error: Optional[RequestError] = None
        self.duration: Optional[float] = None
        self.retry_count = -1

    async def fetch(self, engine: "Engine") -> Optional[Response]:
        loop = asyncio.get_running_loop()
        while self.retry_count < self.max_retry:
            self.retry_count += 1
            start_time = loop.time()
            try:
                self.response = await engine.fetch(request=self)
                self.error = None
            except RequestError as request_error:
                self.error = request_error
            self.duration = loop.time() - start_time
            if self.response:
                return self.response

    def ok(self) -> bool:
        return self.response.ok() if self.response else False

    def result(self) -> str:
        content = self.response and self.response.status or self.error
        if self.state is RequestState.COMPLETED:
            url = truncate_url(self.url, 79, "...")
            return f"{url} {content} {self.duration}"
        elif self.state is RequestState.IN_PROGRESS:
            url = truncate_url(self.url, 77, "...")
            retry = (
                f"({self.error}, retrying {self.retry_count} /" f" {self.max_retry})"
                if self.retry_count > 0
                else ""
            )
            return f"{url} {retry}"
        else:
            return f"{self.url} request not started"

    def __hash__(self) -> int:
        return hash(self.url)

    def __eq__(self, other: Request) -> bool:  # type: ignore
        return self.url == other.url



from typing import Optional


class Response:
    def __init__(
        self,
        status: int,
        content_type: Optional[str] = None,
        text: Optional[str] = None,
    ):
        self.status = status
        self.text = text
        self.content_type = content_type

    def is_html(self) -> bool:
        return not self.content_type or any(
            mime_type in self.content_type
            for mime_type in ["text/html", "application/xhtml+xml"]
        )

    def ok(self) -> bool:
        return self.status < 400



import asyncio
from time import time
from typing import Dict, List, Optional

from rich.console import (
    Console,
    ConsoleOptions,
    ConsoleRenderable,
    Control,
    RenderResult,
)
from rich.control import ControlType
from rich.live import Live
from rich.spinner import Spinner

from .node import Node
from .request import Request, RequestState


class StatusBar(ConsoleRenderable):  # type: ignore
    def __init__(self, in_progress_display: bool = True):
        self.in_progress_display = in_progress_display
        self.start_time = time()
        self.requests_status = {"queued": 0, "success": 0, "error": 0, "completed": 0}
        self.spinner: Spinner = Spinner(name="dots")
        self.completed_nodes: Dict[str, Node] = {}
        self.in_progress_node: Optional[Node] = None

    def notify_request(self, request: Request) -> None:
        if request.state is RequestState.COMPLETED:
            self.requests_status["completed"] += 1
            if request.ok():
                self.requests_status["success"] += 1
            else:
                self.requests_status["error"] += 1

    def notify_node(self, node: Node, children_completed: bool = False) -> None:
        if node.location in self.completed_nodes:
            if children_completed:
                del self.completed_nodes[node.location]
                if node is self.in_progress_node:
                    self.in_progress_node = None
        else:
            self.completed_nodes[node.location] = node

    def elapsed_time(self) -> float:
        return time() - self.start_time

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        if self.in_progress_display:
            if not self.in_progress_node:
                in_progress_nodes: List[Node] = sorted(
                    [
                        node
                        for node in self.completed_nodes.values()
                        if node.count_in_progress() > 0
                    ],
                    key=lambda node: node.count_in_progress(),
                    reverse=True,
                )
                self.in_progress_node = next(iter(in_progress_nodes), None)

            if self.in_progress_node:
                self.spinner.text = self.in_progress_node.get_result_string()
            else:
                self.spinner.text = "pending requests..."
            yield self.spinner

        yield (
            f"Completed: {self.requests_status['completed']} "
            f"| Success: {self.requests_status['success']} "
            f"| Error: {self.requests_status['error']} "
            f"| Elapsed time: {self.elapsed_time():.1f}s"
        )


class Output:
    def __init__(
        self,
        # in_progress_display: int = 3,
        # status_bar: bool = True,
    ):
        self.console = Console()
        self._ready_nodes = {}
        self._completed_requests = {}
        self._status_bar = StatusBar()

    async def start(self):
        try:
            self.console.show_cursor(False)
            with Live(
                console=self.console,
                renderable=self._status_bar,
                auto_refresh=False,
                transient=True,
            ) as live:
                while True:
                    self.display_completed_nodes()
                    live.refresh()
                    await asyncio.sleep(0.05)

        except asyncio.CancelledError:
            raise

        finally:
            self.display_completed_nodes()
            self.console.show_cursor(True)

    def display_completed_nodes(self) -> None:
        displayed_locations = []
        for location, node in self._ready_nodes.items():
            if node.links_all_completed():
                self.console.print(
                    node.get_result_string(),
                    end=f"{Control((ControlType.ERASE_IN_LINE, 0))}\n",
                )
                self._status_bar.notify_node(node, children_completed=True)
                displayed_locations.append(location)
        for location in displayed_locations:
            del self._ready_nodes[location]

    def notify_node(self, node: Node) -> None:
        self._ready_nodes[node.location] = node
        self._status_bar.notify_node(node)

    def notify_request(self, request: Request) -> None:
        self._status_bar.notify_request(request)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['statlink']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'aiolimiter>=1.0.0,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'lxml>=4.9.1,<5.0.0',
 'playwright>=1.28.0,<2.0.0',
 'rich>=12.5.1,<13.0.0']

entry_points = \
{'console_scripts': ['statlink = statlink.command:main']}

setup_kwargs = {
    'name': 'statlink',
    'version': '0.1.4',
    'description': 'An asyncio based multipurpose link checker.',
    'long_description': '# Statlink\n\nAn asyncio based multipurpose link explorer.\n\n- Check broken links in websites.\n- Headless browsers support (Chromium, Firefox, WebKit).\n- Domain / URL Regex / status code filters.\n- Incremental timeout.\n- Panda dataframe / CSV output.\n- Database store support.\n\n\n## Get started\n\n### Installation\n\n    pip install statlink\n\nWith browser support:\n\n    pip install statlink[chromium|fifefox|webkit|all]\n\n## Usage\n\n    Usage: statlink [OPTIONS] [SOURCES]...\n\n    Options:\n      -c, --concurrency INTEGER       Set the maximum number of concurrent\n                                      requests.  [default: 20]\n      --allow-external                Allow external URL to be checked.\n      -d, --depth INTEGER             Recursion depth value for checking URLs,\n                                      negative value will set it to infinity.\n                                      [default: -1]\n      -t, --timeout INTEGER           Set timeout for each requests.  [default:\n                                      20]\n      --engine [aiohttp|chromium|firefox|webkit]\n                                      Engine that will be used to make requests.\n                                      [default: aiohttp]\n      --help                          Show this message and exit.\n\n\n## Examples\n\nFind the dead links for a https domain\n\n    $ statlink https://google.com\n\n\nFind the dead links in a text file\n\n    $ statlink -d 0 bad_urls.txt\n\n\n## Library\n\n\n    from statlink.crawler import Crawler\n\n    crawler = Crawler()\n    crawler.add_url("https://google.com")\n    crawler.start()\n',
    'author': 'Thibaut Frain',
    'author_email': 'thibaut.frain@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Tuple

import click

from .crawler import (
    DEFAULT_CONCURRENCY,
    DEFAULT_DEPTH,
    DEFAULT_ENGINE,
    DEFAULT_MAX_RETRY,
    DEFAULT_RATE_LIMIT,
    DEFAULT_TIMEOUT,
    Crawler,
)


@click.command(no_args_is_help=True)
@click.argument("sources", nargs=-1)
@click.option(
    "--concurrency",
    "-c",
    default=DEFAULT_CONCURRENCY,
    show_default=True,
    type=int,
    help="Set the maximum number of concurrent requests.",
)
@click.option(
    "--recursion-mode",
    "-m",
    default="internal",
    show_default=True,
    type=click.Choice(["internal", "all"]),
    help="Set the recursion mode for the crawler.",
)
@click.option(
    "--check-mode",
    default="all",
    show_default=True,
    type=click.Choice(["internal", "all"]),
)
@click.option(
    "--depth",
    "-d",
    default=DEFAULT_DEPTH,
    show_default=True,
    type=int,
    help="Recursion depth value for checking URLs, "
    "negative value will set it to infinity.",
)
@click.option(
    "--timeout",
    "-t",
    default=DEFAULT_TIMEOUT,
    show_default=True,
    type=int,
    help="Set timeout for each requests.",
)
@click.option(
    "--max-retry",
    "-r",
    default=DEFAULT_MAX_RETRY,
    show_default=True,
    type=int,
    help="Set the maximum number of time a request will be retried if failed",
)
@click.option(
    "--engine",
    default=DEFAULT_ENGINE,
    show_default=True,
    type=click.Choice(["aiohttp", "chromium", "firefox", "webkit"]),
    help="Engine that will be used to make requests.",
)
@click.option(
    "--rate-limit",
    "-l",
    default=DEFAULT_RATE_LIMIT,
    show_default=True,
    type=int,
    help="Set the maximum number of requests made per second.",
)
def main(
    sources: Tuple[str],
    concurrency: int,
    recursion_mode: str,
    check_mode: str,
    depth: int,
    timeout: int,
    max_retry: int,
    engine: str,
    rate_limit: int,
) -> None:

    Crawler(
        concurrency=concurrency,
        recursion_mode=recursion_mode,
        check_mode=check_mode,
        depth=depth,
        timeout=timeout,
        max_retry=max_retry,
        engine=engine,
        rate_limit=rate_limit,
    ).start(sources=sources)



from __future__ import annotations

import asyncio
import subprocess
from abc import abstractmethod
from contextlib import AbstractAsyncContextManager
from functools import partial
from typing import Dict, Optional, Type

import aiohttp
from playwright import async_api as playwright
from playwright._impl._driver import compute_driver_executable, get_driver_env

from .request import Request, RequestConnectionError, RequestError, RequestTimeoutError
from .response import Response


class Engine(AbstractAsyncContextManager):  # type: ignore
    @abstractmethod
    async def fetch(self, request: "Request") -> Optional[Response]:
        pass  # pragma: no cover


class AioHTTPEngine(Engine):
    def __init__(self):
        conn = aiohttp.TCPConnector(limit_per_host=20, limit=30)
        self.client_session = aiohttp.ClientSession(connector=conn)

    async def __aenter__(self) -> AioHTTPEngine:
        await self.client_session.__aenter__()
        return self

    async def __aexit__(self, *exc):
        await self.client_session.__aexit__(*exc)

    async def fetch(self, request: "Request") -> Optional[Response]:
        try:
            async with self.client_session.get(
                request.url, timeout=request.timeout
            ) as client_response:
                response = Response(
                    status=client_response.status,
                    content_type=client_response.headers.get("Content-Type"),
                )
                if response.is_html():
                    try:
                        response.text = await client_response.text()
                    except ValueError:
                        pass
                if not response.text:
                    client_response.close()
                return response
        except asyncio.TimeoutError:
            raise RequestTimeoutError()
        except aiohttp.ClientConnectionError as exc:
            raise RequestConnectionError(exc)
        except asyncio.CancelledError:
            raise
        except Exception as exc:
            raise RequestError(exc)


class BrowserEngine(Engine):
    def __init__(self, browser_type: str):
        self.playwright_context = playwright.async_playwright()
        self.browser_type = browser_type
        self.browser: playwright.Browser

    async def __aenter__(self) -> BrowserEngine:
        async_playwright = await self.playwright_context.__aenter__()
        try:
            self.browser = await self.launch_browser(async_playwright)
        except playwright.Error as exc:
            if "Executable doesn't exist" in str(exc):
                install_browser_executable(self.browser_type)
                self.browser = await self.launch_browser(async_playwright)
            else:
                raise
        return self

    async def __aexit__(self, *exc):
        await self.playwright_context.__aexit__(*exc)

    async def launch_browser(
        self, async_playwright: playwright.Playwright
    ) -> playwright.Browser:
        browser: playwright.Browser = await getattr(
            async_playwright, self.browser_type
        ).launch()
        return browser

    async def fetch(self, request: "Request") -> Optional[Response]:
        page = await self.browser.new_page()
        try:
            response = await page.goto(
                request.url,
                timeout=request.timeout * 1000,
                wait_until="domcontentloaded",
            )

            if response:
                return Response(
                    text=await page.content(),
                    status=response.status,
                    content_type=response.headers.get("Content-Type"),
                )

        except playwright.TimeoutError:
            raise RequestTimeoutError()

        except playwright.Error as exc:
            error_message = str(exc).lower()
            connection_error_messages = [
                "err_name_not_resolved",
                "err_address_unreachable",
            ]
            if any(message in error_message for message in connection_error_messages):
                raise RequestConnectionError()
            raise RequestError(self._truncate_error(str(exc)))

        except asyncio.CancelledError:
            raise

        except Exception as exc:
            raise RequestError(exc)

        finally:
            await page.close()

    @staticmethod
    def _truncate_error(error_message: str) -> str:
        return error_message.split("\n=", maxsplit=1)[0].capitalize()


ENGINE_MAPPING: Dict[str, Type[Engine]] = {
    "aiohttp": AioHTTPEngine,
    **{
        browser_type: partial(BrowserEngine, browser_type=browser_type)  # type: ignore
        for browser_type in ["chromium", "firefox", "webkit"]
    },
}


def install_browser_executable(browser_type: str) -> None:
    driver_executable = compute_driver_executable()
    subprocess.run(
        [str(driver_executable), "install", browser_type], env=get_driver_env()
    )
